//! [Monad] for passing and combining metadata related to the evaluation of values.
//!
//! For "either"/"sum" wrapping, see [`result`].
//!
//! For no extra data, see [`solo`].
//!
//! [`result`]: super::result
//! [`solo`]: super::solo

use crate::func::class_prelude::*;

/// Metadata type.
pub trait Effect: Send {
    /// Used in [`Pure::pure`].
    fn e_pure() -> Self;

    /// Used in [`ApplicativeLA2::la2`] and other [`Applicative`] methods.
    fn e_parallel(el: Self, er: Self) -> Self;

    /// Used in [`Monad::bind`] and other [`Monad`] methods.
    fn e_after(self, effect: Self) -> Self;
}

/// Value and related metadata.
#[derive(Clone)]
pub struct WithEffect<A, E> {
    pub value: A,
    pub effect: E,
}

impl<A, E: Effect> WithEffect<A, E> {
    fn e_after(self, effect: E) -> Self {
        WithEffect {
            value: self.value,
            effect: self.effect.e_after(effect),
        }
    }
}

#[derive(SharedFunctorAny)]
pub struct EffectInstance<E: Send>(E);

impl<E: Send> WeakFunctorAny for EffectInstance<E> {
    type FAny<'a, A: 'a + Send> = WithEffect<A, E>
    where
        Self: 'a;
}

impl<'a, E: 'a + Send> Functor<'a> for EffectInstance<E> {
    fn fmap<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, f: impl 'a + FnOnce(A) -> B) -> Self::F<B> {
        WithEffect {
            value: f(fa.value),
            effect: fa.effect,
        }
    }

    fn replace<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, b: B) -> Self::F<B> {
        drop(fa.value);
        WithEffect {
            value: b,
            effect: fa.effect,
        }
    }
}

impl<'a, E: 'a + Effect> Pure<'a> for EffectInstance<E> {
    fn pure<A: 'a + Send>(a: A) -> Self::F<A> {
        WithEffect {
            value: a,
            effect: E::e_pure(),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeSeq<'a> for EffectInstance<E> {
    fn seq<A: 'a + Send, B: 'a + Send>(
        ff: Self::F<impl 'a + Send + FnOnce(A) -> B>,
        fa: Self::F<A>,
    ) -> Self::F<B> {
        WithEffect {
            value: (ff.value)(fa.value),
            effect: E::e_parallel(ff.effect, fa.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeLA2<'a> for EffectInstance<E> {
    fn la2<A: 'a + Send, B: 'a + Send, C: 'a + Send>(
        fa: Self::F<A>,
        fb: Self::F<B>,
        f: impl 'a + Send + FnOnce(A, B) -> C,
    ) -> Self::F<C> {
        WithEffect {
            value: f(fa.value, fb.value),
            effect: E::e_parallel(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeTuple<'a> for EffectInstance<E> {
    fn tuple<A: 'a + Send, B: 'a + Send>((fa, fb): (Self::F<A>, Self::F<B>)) -> Self::F<(A, B)> {
        WithEffect {
            value: (fa.value, fb.value),
            effect: E::e_parallel(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> ApplicativeSelect<'a> for EffectInstance<E> {}

impl<'a, E: 'a + Effect> Applicative<'a> for EffectInstance<E> {
    fn discard_first<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<B> {
        drop(fa.value);
        WithEffect {
            value: fb.value,
            effect: E::e_parallel(fa.effect, fb.effect),
        }
    }

    fn discard_second<A: 'a + Send, B: 'a + Send>(fa: Self::F<A>, fb: Self::F<B>) -> Self::F<A> {
        drop(fb.value);
        WithEffect {
            value: fa.value,
            effect: E::e_parallel(fa.effect, fb.effect),
        }
    }
}

impl<'a, E: 'a + Effect> Monad<'a> for EffectInstance<E> {
    fn bind<A: 'a + Send, B: 'a + Send>(
        fa: Self::F<A>,
        f: impl 'a + FnOnce(A) -> Self::F<B>,
    ) -> Self::F<B> {
        f(fa.value).e_after(fa.effect)
    }

    fn iterate<B: 'a + Send>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::F<B> {
        let mut effect = E::e_pure();
        loop {
            let fa = f.next();
            effect = fa.effect.e_after(effect);
            match fa.value {
                ControlFlow::Continue(next_f) => f = next_f,
                ControlFlow::Break(b) => return WithEffect { value: b, effect },
            }
        }
    }

    fn join<A: 'a + Send>(ffa: Self::F<Self::F<A>>) -> Self::F<A> {
        ffa.value.e_after(ffa.effect)
    }
}

impl<'a, E: 'a + Send> LocalFunctor<'a> for EffectInstance<E> {
    fn stuff<A: 'a + Send, T: Pure<'a>>(fa: Self::F<T::F<A>>) -> T::F<Self::F<A>> {
        T::fmap(fa.value, |a| WithEffect {
            value: a,
            effect: fa.effect,
        })
    }
}

#[cfg(test)]
mod effect_tests {
    use std::sync::Arc;

    use super::{test_suite, tests, Effect, EffectInstance, WithEffect};

    #[derive(Clone, Debug, PartialEq)]
    enum TestEffect {
        Pure,
        Sample(i32),
        Parallel(Arc<Self>, Arc<Self>),
        Sequential(Arc<Self>, Arc<Self>),
    }

    impl TestEffect {
        fn is_pure(&self) -> bool {
            matches!(self, Self::Pure)
        }

        fn simplify(&self) -> Arc<Self> {
            match self {
                Self::Pure => Self::Pure.into(),
                Self::Sample(x) => Self::Sample(*x).into(),
                Self::Parallel(a, b) if a.is_pure() => b.clone(),
                Self::Parallel(a, b) if b.is_pure() => a.clone(),
                Self::Sequential(a, b) if a.is_pure() => b.clone(),
                Self::Sequential(a, b) if b.is_pure() => a.clone(),
                Self::Parallel(a, b) => match a.as_ref() {
                    Self::Parallel(aa, ab) => {
                        Self::Parallel(aa.clone(), Self::Parallel(ab.clone(), b.clone()).into())
                            .simplify()
                    }
                    a => Self::Parallel(a.simplify(), b.simplify()).into(),
                },
                Self::Sequential(a, b) => match a.as_ref() {
                    TestEffect::Sequential(aa, ab) => {
                        Self::Sequential(aa.clone(), Self::Sequential(ab.clone(), b.clone()).into())
                            .simplify()
                    }
                    a => Self::Sequential(a.simplify(), b.simplify()).into(),
                },
            }
        }
    }

    impl Effect for TestEffect {
        fn e_pure() -> Self {
            Self::Pure
        }

        fn e_parallel(el: Self, er: Self) -> Self {
            Self::Parallel(el.into(), er.into())
        }

        fn e_after(self, effect: Self) -> Self {
            Self::Sequential(effect.into(), self.into())
        }
    }

    type T = EffectInstance<TestEffect>;

    impl<'a> tests::Eqr<'a> for T {
        fn eqr<A: 'a + Send + PartialEq + std::fmt::Debug>(
            name: &'a str,
            left: Self::F<A>,
            right: Self::F<A>,
        ) -> tests::R {
            tests::eqr(
                name,
                (left.value, left.effect.simplify()),
                (right.value, right.effect.simplify()),
            )
        }
    }

    impl<'a> test_suite::FunctorTestSuite<'a> for T {
        fn sample<A: 'a + Send, F: FnMut(Arc<dyn test_suite::WrapFunction<'a, A, Self::F<A>>>)>(
            mut f: F,
        ) {
            f(Arc::new(|a| WithEffect {
                value: a,
                effect: TestEffect::Sample(0),
            }));
            f(Arc::new(|a| WithEffect {
                value: a,
                effect: TestEffect::Sample(1),
            }));
            f(Arc::new(|a| WithEffect {
                value: a,
                effect: TestEffect::Sample(2),
            }));
            f(Arc::new(|a| WithEffect {
                value: a,
                effect: TestEffect::Sample(3),
            }));
        }
    }

    #[test]
    fn monad_follows_laws() {
        test_suite::monad_follows_laws::<T>().unwrap();
    }

    #[test]
    fn shared_follows_laws() {
        test_suite::shared_follows_laws::<T>().unwrap();
    }

    #[test]
    fn local_follows_laws() {
        test_suite::local_follows_laws::<T>().unwrap();
    }
}
